@@ -1,4 +1,5 @@
-from testify import TestCase, run, test_discovery
+from functools import wraps
+from testify import TestCase, run, test_discovery, assert_length
 from os.path import dirname, join, abspath
 from os import getcwd, chdir
 
@@ -11,6 +12,7 @@ class DiscoveryTestCase(TestCase):
 
 def relative(func):
     'decorator for tests that rely on relative paths'
+    @wraps(func)
     def wrapped(*args, **kwargs):
         cwd = getcwd()
         chdir(HERE)
@@ -41,5 +43,15 @@ class TestDiscoverFilePath(DiscoveryTestCase):
         self.discover(join(HERE, 'subdir/test.py'))
 
 
+class TestDiscoverIgnoreImportedThings(DiscoveryTestCase):
+    def test_imported_things_are_ignored(self):
+        #TODO CHANGE MY NAME
+        discovered_imported = list(test_discovery.discover('test.test_suite_subdir.import_testcase'))
+        discovered_actually_defined_in_module = list(test_discovery.discover('test.test_suite_subdir.define_testcase'))
+
+        assert_length(discovered_imported, 0)
+        assert_length(discovered_actually_defined_in_module, 1)
+
+
 if __name__ == '__main__':
     run()
