@@ -17,10 +17,13 @@
 
 ### TODO: finish doing the retry stuff for the inner clauses
 
+from __future__ import with_statement
+
 __author__ = "Oliver Nicholas <bigo@yelp.com>"
 __testify = 1
 
 from collections import defaultdict
+from contextlib import contextmanager
 import inspect
 from new import instancemethod
 import sys
@@ -31,7 +34,7 @@ import deprecated_assertions
 from testify.utils import class_logger
 
 # just a useful list to have
-fixture_types = ['class_setup', 'setup', 'teardown', 'class_teardown']
+fixture_types = ['class_setup', 'setup', 'teardown', 'class_teardown', 'setup_teardown', 'class_setup_teardown']
 deprecated_fixture_type_map = {
     'classSetUp': 'class_setup',
     'setUp': 'setup',
@@ -121,11 +124,22 @@ class TestCase(object):
     STAGE_TEARDOWN = 4
     STAGE_CLASS_TEARDOWN = 5
 
+    EVENT_ON_RUN_TEST_METHOD = 1
+    EVENT_ON_COMPLETE_TEST_METHOD = 2
+    EVENT_ON_RUN_CLASS_SETUP_METHOD = 3
+    EVENT_ON_COMPLETE_CLASS_SETUP_METHOD = 4
+    EVENT_ON_RUN_CLASS_TEARDOWN_METHOD = 5
+    EVENT_ON_COMPLETE_CLASS_TEARDOWN_METHOD = 6
+    EVENT_ON_RUN_FIXTURE_METHOD = 7
+    EVENT_ON_COMPLETE_FIXTURE_METHOD = 8
+
     log = class_logger.ClassLogger()
 
     def __init__(self, *args, **kwargs):
         super(TestCase, self).__init__()
 
+        self._method_level = False
+
         # ascend the class hierarchy and discover fixture methods
         self.__init_fixture_methods()
 
@@ -134,18 +148,10 @@ class TestCase(object):
         self.__suites_require = kwargs.get('suites_require', set())
         self.__name_overrides = kwargs.get('name_overrides', None)
 
-        # if the class has any suites applied to it, copy them down into its test methods
-        if hasattr(self, '_suites'):
-            for member_name in dir(self):
-                if member_name.startswith('test'):
-                    member = getattr(self, member_name)
-                    if isinstance(member, types.MethodType):
-                        suited_function = suite(*getattr(self, '_suites'))(member)
-                        setattr(self, member_name, suited_function)
+        self.__debugger = kwargs.get('debugger')
 
         # callbacks for various stages of execution, used for stuff like logging
-        self.__on_run_test_method_callbacks = []
-        self.__on_complete_test_method_callbacks = []
+        self.__callbacks = defaultdict(list)
 
         # one of these will later be populated with exception info if there's an
         # exception in the class_setup/class_teardown stage
@@ -161,6 +167,9 @@ class TestCase(object):
             if name.startswith(('assert', 'fail')):
                 setattr(self, name, instancemethod(getattr(deprecated_assertions, name), self, self.__class__))
 
+        self.failure_limit = kwargs.pop('failure_limit', None)
+        self.failure_count = 0
+
     def __init_fixture_methods(self):
         """Initialize and populate the lists of fixture methods for this TestCase.
         Fixture methods are added by the MetaTestCase metaclass at runtime only to
@@ -202,76 +211,72 @@ class TestCase(object):
         limit itself to test methods in those suites.
         """
         for member_name in dir(self):
-            if member_name.startswith("test"):
-                member = getattr(self, member_name)
-                if inspect.ismethod(member):
-                    member_suites = set(getattr(member, '_suites', set()))
-                    # if there are any exclude suites, exclude methods under them
-                    if (not self.__suites_exclude) or (not self.__suites_exclude & member_suites):
-                        # if there are any include suites, only run methods in them
-                        if not self.__suites_include or (self.__suites_include & member_suites):
-                            # if there are any require suites, only run methods in *all* of those suites
-                            if not self.__suites_require or ((self.__suites_require & member_suites) == self.__suites_require):
-                                # if there are any name overrides, only run the named methods
-                                if self.__name_overrides is None or member.__name__ in self.__name_overrides:
-                                    yield member
+            if not member_name.startswith("test"):
+                continue
+            member = getattr(self, member_name)
+            if not inspect.ismethod(member):
+                continue
+            member_suites = getattr(member, '_suites', set()) | set(getattr(self, '_suites', []))
+            # if there are any exclude suites, exclude methods under them
+            if self.__suites_exclude and self.__suites_exclude & member_suites:
+                continue
+            # if there are any include suites, only run methods in them
+            if self.__suites_include and not (self.__suites_include & member_suites):
+                continue
+            # if there are any require suites, only run methods in *all* of those suites
+            if self.__suites_require and not ((self.__suites_require & member_suites) == self.__suites_require):
+                continue
+
+            # if there are any name overrides, only run the named methods
+            if self.__name_overrides is None or member.__name__ in self.__name_overrides:
+                yield member
 
     def run(self):
         """Delegator method encapsulating the flow for executing a TestCase instance"""
+
         self.__run_class_setup_fixtures()
-        self.__run_test_methods()
+        self.__enter_context_managers(self.class_setup_teardown_fixtures, self.__run_test_methods)
         self.__run_class_teardown_fixtures()
 
     def __run_class_setup_fixtures(self):
         """Running the class's class_setup method chain."""
-        self._stage = self.STAGE_CLASS_SETUP
-
-        for fixture_method in self.class_setup_fixtures:
-            result = TestResult(fixture_method)
-
-            try:
-                for callback in self.__on_run_test_method_callbacks:
-                    callback(self, fixture_method)
-
-                result.start()
-
-                if self.__execute_block_recording_exceptions(fixture_method, result, is_class_level=True):
-                    result.end_in_success()
-            except (KeyboardInterrupt, SystemExit):
-                result.end_in_incomplete(sys.exc_info())
-                for callback in self.__on_complete_test_method_callbacks:
-                    callback(self, result)
-                raise
-            else:
-                for callback in self.__on_complete_test_method_callbacks:
-                    callback(self, result)
-
-        self.__run_deprecated_fixture_method('classSetUp')
+        self.__run_class_fixtures(
+            self.STAGE_CLASS_SETUP,
+            self.class_setup_fixtures + [ self.classSetUp ],
+            self.EVENT_ON_RUN_CLASS_SETUP_METHOD,
+            self.EVENT_ON_COMPLETE_CLASS_SETUP_METHOD,
+        )
 
     def __run_class_teardown_fixtures(self):
         """End the process of running tests.  Run the class's class_teardown methods"""
-        self._stage = self.STAGE_CLASS_TEARDOWN
+        self.__run_class_fixtures(
+            self.STAGE_CLASS_TEARDOWN,
+            [ self.classTearDown ] + self.class_teardown_fixtures,
+            self.EVENT_ON_RUN_CLASS_TEARDOWN_METHOD,
+            self.EVENT_ON_COMPLETE_CLASS_TEARDOWN_METHOD,
+        )
 
-        self.__run_deprecated_fixture_method('classTearDown')
+    def __run_class_fixtures(self, stage, fixtures, callback_on_run_event, callback_on_complete_event):
+        """Set the current _stage, run a set of fixtures, calling callbacks before and after each."""
+        self._stage = stage
 
-        for fixture_method in self.class_teardown_fixtures:
+        for fixture_method in fixtures:
             result = TestResult(fixture_method)
+
             try:
-                for callback in self.__on_run_test_method_callbacks:
-                    callback(self, fixture_method)
+                for callback in self.__callbacks[callback_on_run_event]:
+                    callback(result.to_dict())
 
                 result.start()
 
                 if self.__execute_block_recording_exceptions(fixture_method, result, is_class_level=True):
                     result.end_in_success()
             except (KeyboardInterrupt, SystemExit):
-                result.end_in_incomplete(sys.exc_info())
-                for callback in self.__on_complete_test_method_callbacks:
-                    callback(self, result)
+                result.end_in_interruption(sys.exc_info())
                 raise
-            else:
-                for callback in self.__on_complete_test_method_callbacks:
-                    callback(self, result)
+            finally:
+                for callback in self.__callbacks[callback_on_complete_event]:
+                    callback(result.to_dict())
 
     @classmethod
     def in_suite(cls, method, suite_name):
@@ -287,6 +292,14 @@ class TestCase(object):
         method_suites = set(getattr(method, '_suites', set()))
         return (self.__suites_exclude & method_suites)
 
+    def __enter_context_managers(self, fixture_methods, callback):
+        """Transform each fixture_method into a context manager with contextlib.contextmanager, enter them recursively, and call callback"""
+        if fixture_methods:
+            with contextmanager(fixture_methods[0])():
+                self.__enter_context_managers(fixture_methods[1:], callback)
+        else:
+            callback()
+
     def __run_test_methods(self):
         """Run this class's setup fixtures / test methods / teardown fixtures.
 
@@ -304,9 +317,11 @@ class TestCase(object):
             test_method.im_self.test_result = result
 
             try:
+                self._method_level = True # Flag that we're currently running method-level stuff (rather than class-level)
+
                 # run "on-run" callbacks. eg/ print out the test method name
-                for callback in self.__on_run_test_method_callbacks:
-                    callback(self, test_method)
+                for callback in self.__callbacks[self.EVENT_ON_RUN_TEST_METHOD]:
+                    callback(result.to_dict())
                 result.start()
 
                 if self.__class_level_failure:
@@ -317,21 +332,27 @@ class TestCase(object):
                     # first, run setup fixtures
                     self._stage = self.STAGE_SETUP
                     def _setup_block():
-                        for fixture_method in self.setup_fixtures:
+                        for fixture_method in self.setup_fixtures + [ self.setUp ]:
                             fixture_method()
-                        self.__run_deprecated_fixture_method('setUp')
                     self.__execute_block_recording_exceptions(_setup_block, result)
 
-                    # then run the test method itself, assuming setup was successful
-                    self._stage = self.STAGE_TEST_METHOD
+                    def _run_test_block():
+                        # then run the test method itself, assuming setup was successful
+                        self._stage = self.STAGE_TEST_METHOD
+                        if not result.complete:
+                            self.__execute_block_recording_exceptions(test_method, result)
+
+                    def _setup_teardown_block():
+                        self.__enter_context_managers(self.setup_teardown_fixtures, _run_test_block)
+
+                    # then run any setup_teardown fixtures, assuming setup was successful.
                     if not result.complete:
-                        self.__execute_block_recording_exceptions(test_method, result)
+                        self.__execute_block_recording_exceptions(_setup_teardown_block, result)
 
                     # finally, run the teardown phase
                     self._stage = self.STAGE_TEARDOWN
                     def _teardown_block():
-                        self.__run_deprecated_fixture_method('tearDown')
-                        for fixture_method in self.teardown_fixtures:
+                        for fixture_method in [ self.tearDown ] + self.teardown_fixtures:
                             fixture_method()
                     self.__execute_block_recording_exceptions(_teardown_block, result)
 
@@ -339,16 +360,18 @@ class TestCase(object):
                 if not result.complete:
                     result.end_in_success()
             except (KeyboardInterrupt, SystemExit):
-                result.end_in_incomplete(sys.exc_info())
-                for callback in self.__on_complete_test_method_callbacks:
-                    callback(self, result)
+                result.end_in_interruption(sys.exc_info())
                 raise
-            else:
-                for callback in self.__on_complete_test_method_callbacks:
-                    callback(self, result)
+            finally:
+                for callback in self.__callbacks[self.EVENT_ON_COMPLETE_TEST_METHOD]:
+                    callback(result.to_dict())
 
-    EVENT_ON_RUN_TEST_METHOD = 1
-    EVENT_ON_COMPLETE_TEST_METHOD = 2
+                self._method_level = False
+
+                if not result.success:
+                    self.failure_count += 1
+                    if self.failure_limit and self.failure_count >= self.failure_limit:
+                        return
 
     def register_callback(self, event, callback):
         """Register a callback for an internal event, usually used for logging.
@@ -357,12 +380,7 @@ class TestCase(object):
 
         Fixture objects can be distinguished by the running them through self.is_fixture_method().
         """
-        if event == self.EVENT_ON_RUN_TEST_METHOD:
-            self.__on_run_test_method_callbacks.append(callback)
-        elif event == self.EVENT_ON_COMPLETE_TEST_METHOD:
-            self.__on_complete_test_method_callbacks.append(callback)
-        else:
-            raise ValueError("Invalid callback event: %s" % event)
+        self.__callbacks[event].append(callback)
 
     def __execute_block_recording_exceptions(self, block_fxn, result, is_class_level=False):
         """Excerpted code for executing a block of code that might except and cause us to update a result object.
@@ -399,6 +417,12 @@ class TestCase(object):
                 result.end_in_error(exc_info)
                 if is_class_level:
                     self.__class_level_error = exc_info
+            if self.__debugger:
+                exc, val, tb = exc_info
+                print "\nDEBUGGER"
+                print "\n".join(result.format_exception_info())
+                import ipdb
+                ipdb.post_mortem(tb)
             return False
         else:
             return True
@@ -415,31 +439,6 @@ class TestCase(object):
             else:
                 return True
 
-    def __run_deprecated_fixture_method(self, fixture_name):
-        """This runs an old-style (eg/ 'def setUp') fixture method."""
-        if hasattr(self, fixture_name):
-            deprecated_method = getattr(self, fixture_name)
-
-            if fixture_name.startswith('class'):
-                result = TestResult(deprecated_method)
-                try:
-                    for callback in self.__on_run_test_method_callbacks:
-                        callback(self, deprecated_method)
-
-                    result.start()
-                    if self.__execute_block_recording_exceptions(deprecated_method, result, is_class_level=True):
-                        result.end_in_success()
-                except (KeyboardInterrupt, SystemExit):
-                    result.end_in_incomplete(sys.exc_info())
-                    for callback in self.__on_complete_test_method_callbacks:
-                        callback(self, result)
-                    raise
-                else:
-                    for callback in self.__on_complete_test_method_callbacks:
-                        callback(self, result)
-            else:
-                deprecated_method()
-
 def suite(*args, **kwargs):
     """Decorator to conditionally assign suites to individual test methods.
 
@@ -476,3 +475,36 @@ class_setup = __fixture_decorator_factory('class_setup')
 setup = __fixture_decorator_factory('setup')
 teardown = __fixture_decorator_factory('teardown')
 class_teardown = __fixture_decorator_factory('class_teardown')
+
+setup_teardown = __fixture_decorator_factory('setup_teardown')
+class_setup_teardown = __fixture_decorator_factory('class_setup_teardown')
+
+class let(object):
+    """Decorator that creates a lazy-evaluated helper property. The value is
+    cached across multiple calls in the same test, but not across multiple
+    tests.
+    """
+
+    class _unsaved(object): pass
+
+    def __init__(self, func):
+        self._func = func
+        self._result = self._unsaved
+
+    def __get__(self, test_case, cls):
+        if test_case is None:
+            return self
+        if self._result is self._unsaved:
+            self._save_result(self._func(test_case))
+            self._register_reset_after_test_completion(test_case)
+        return self._result
+
+    def _save_result(self, result):
+        self._result = result
+
+    def _register_reset_after_test_completion(self, test_case):
+        test_case.register_callback(TestCase.EVENT_ON_COMPLETE_TEST_METHOD,
+                                    lambda _: self._reset_value())
+
+    def _reset_value(self):
+        self._result = self._unsaved
