@@ -26,12 +26,16 @@ class TestReporter(object):
         """
         self.options = options
 
-    def test_start(self, test_case, method):
-        """Called when a test method is being run."""
+    def test_counts(self, test_case_count, test_method_count):
+        """Called after discovery finishes. May not be called by all test runners, e.g. TestRunnerClient."""
         pass
 
-    def test_complete(self, test_case, result):
-        """Called when a test case is complete"""
+    def test_start(self, result):
+        """Called when a test method is being run. Gets passed a TestResult dict which should not be complete."""
+        pass
+
+    def test_complete(self, result):
+        """Called when a test method is complete. result is a TestResult dict which should be complete."""
         pass
 
     def report(self):
